/*   Copyright (C) 1994, Nienhaus & Associates, Inc.
	U. S. Government Sponsorship under NASA Contract NAS9-19039
	is acknowledged.
*/

//=======================================================
//   SndMsg.cc
//=======================================================
#include "SndMsg.h"
#include "Application.h"
#include "twW.h"
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/BulletinB.h>
#include <Xm/Label.h>
#include <Xm/Separator.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>


void traversal( Widget, XtPointer, XtPointer );
void checkNode( Widget, XtPointer, XtPointer );
void valueChanged(Widget, XtPointer, XtPointer ); 

SndMsgCmd::SndMsgCmd ( char          *name, 
	                  int            active):
	                  NoUndoCmd ( name, active )
{ 
	strcpy ( _sender, "ping" );
	strcpy ( _receiver, "pong" );

// create the dialog 
	XmString motif_string = XmStringCreateLtoR ( "Send Message",
		XmSTRING_DEFAULT_CHARSET );

	_dialog = XmCreateBulletinBoardDialog ( theApplication->baseWidget(),
		name, NULL, 0 );

	XtVaSetValues  ( _dialog, XmNautoUnmanage, False,
		XmNnoResize, True, XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
		XmNdialogTitle, motif_string, XmNwidth, 350, XmNheight, 300, NULL );

	XmStringFree( motif_string );

// sender
	Widget sender_label = XtVaCreateManagedWidget ( "Sender:",
		xmLabelWidgetClass, _dialog, XmNx, 60, XmNy, 12, NULL );

    _sender_text = XtVaCreateManagedWidget ( "sender_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 10, NULL );

    XmTextSetString ( _sender_text, _sender );

    XtAddCallback( _sender_text, XmNactivateCallback, traversal, NULL );

// send time
	Widget send_time_label = XtVaCreateManagedWidget ( "Send Time:",
		xmLabelWidgetClass, _dialog, XmNx, 42, XmNy, 52, NULL );
                                     
	_send_time_text = XtVaCreateManagedWidget ( "send_time_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 50, NULL );

    XmTextSetString ( _send_time_text, "0" );

    XtAddCallback( _send_time_text, XmNactivateCallback, checkNode, NULL );

    XtAddCallback( _send_time_text, XmNmodifyVerifyCallback, checkNode, NULL );

// receiver
	Widget receiver_label = XtVaCreateManagedWidget ( "Receiver:",
		xmLabelWidgetClass, _dialog, XmNx, 47, XmNy, 92, NULL );

    _receiver_text = XtVaCreateManagedWidget ( "receiver_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 90, NULL );

    XmTextSetString ( _receiver_text, _receiver );

    XtAddCallback( _receiver_text, XmNactivateCallback, traversal, NULL );

// receive time
	Widget receive_time_label = XtVaCreateManagedWidget ( "Receive Time:",
		xmLabelWidgetClass, _dialog, XmNx, 24, XmNy, 132, NULL );
                                     
	_receive_time_text = XtVaCreateManagedWidget ( "receive_time_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 130, NULL );

    XmTextSetString ( _receive_time_text, "0" );

    XtAddCallback( _receive_time_text, XmNactivateCallback, checkNode, NULL );

    XtAddCallback( _receive_time_text, XmNmodifyVerifyCallback,checkNode, NULL);

// selector
    Widget selector_label = XtVaCreateManagedWidget ( "Selector:",
		xmLabelWidgetClass, _dialog, XmNx, 47, XmNy, 172, NULL );
                                     
	_selector_text = XtVaCreateManagedWidget ( "selector_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 170, NULL );

    XmTextSetString ( _selector_text, "0" );
    
    XtAddCallback( _selector_text, XmNmodifyVerifyCallback, checkNode, NULL );

// text
    Widget text_label = XtVaCreateManagedWidget ( "Message Text:",
		xmLabelWidgetClass, _dialog, XmNx, 24, XmNy, 212, NULL );
                                     
	_message_text = XtVaCreateManagedWidget ( "message_text",
		xmTextWidgetClass, _dialog, XmNtraversalOn, TRUE,
		XmNhighlightOnEnter, TRUE, XmNx, 115, XmNy, 210, NULL );

//separator
	Widget sep = XtVaCreateManagedWidget ( "sep", xmSeparatorWidgetClass,
		_dialog, XmNx, 0, XmNy, 250, XmNwidth, 330, NULL );


	Widget form = XtVaCreateManagedWidget ( "row", xmFormWidgetClass,
		_dialog, XmNx, 0, XmNy, 260, NULL );

	Widget _okButton = XtVaCreateManagedWidget ( "   OK   ",
		xmPushButtonWidgetClass, form, XmNrightAttachment, XmATTACH_POSITION,
		XmNrightPosition, 50, NULL );

	XtAddCallback ( _okButton, XmNactivateCallback, &SndMsgCmd::okCallback, 
		(XtPointer) this );

	Widget _cancelButton = XtVaCreateManagedWidget ( "CANCEL",
		xmPushButtonWidgetClass, form, XmNleftAttachment, XmATTACH_POSITION,
		XmNleftPosition, 80, NULL );

	XtAddCallback ( _cancelButton, XmNactivateCallback,
		&SndMsgCmd::cancelCallback, (XtPointer) this );

}


SndMsgCmd::~SndMsgCmd()
{
	XtDestroyWidget( _dialog );
}


void SndMsgCmd::execute()
{
	XtManageChild ( _dialog );
}

extern "C"
{
	VTime newVTime ( STime, Ulong, Ulong );
	int cm_create_event_message ( Name, Name, VTime, VTime, Long, int,
		Message * );
	cm_queue_event_message();
}

void SndMsgCmd::doit()
{
	char * buf, *p;
	double St;
	Ulong S1, S2;

	XtUnmanageChild( _dialog );
	XmUpdateDisplay(theApplication->baseWidget());  
	strncpy ( _sender, XmTextGetString( _sender_text ), NOBJLEN );

	St = 0.0; S1 = S2 = 0;
	p = buf = XmTextGetString( _send_time_text );
	while ( *p )
	{
		if ( *p == ',' )
			*p = ' ';
		p++;
	}
	sscanf ( buf, "%lf%d%d", &St, &S1, &S2 );
	_send_time = newVTime ( St, S1, S2 );

	printf ( "MsgSnd\n" );
	printf ( "Sender: %s\n", _sender );
	printf ( "Send Time: %.4f, %d, %d\n", St, S1, S2 );

	strncpy ( _receiver, XmTextGetString( _receiver_text ), NOBJLEN );
	St = 0.0; S1 = S2 = 0;
	p = buf = XmTextGetString( _receive_time_text );
	while ( *p )
	{
		if ( *p == ',' )
			*p = ' ';
		p++;
	}
	sscanf ( buf, "%lf%d%d", &St, &S1, &S2 );
	_receive_time = newVTime ( St, S1, S2 );

	printf ( "Receiver: %s\n", _receiver );
	printf ( "Receiver Time: %.4f, %d, %d\n", St, S1, S2 );

	_selector = 0;
	sscanf ( XmTextGetString( _selector_text), "%d", &_selector );
	printf ( "Selector: %d\n", _selector );

	_text_length = strlen ( XmTextGetString ( _message_text ) );

	printf ( "Msg (%d bytes): %s\n", _text_length,
		XmTextGetString ( _message_text ) );

	if ( SUCCESS == cm_create_event_message ( _sender, _receiver,
			_send_time, _receive_time,
			_selector, _text_length, XmTextGetString ( _message_text ) ) )
	{
		cm_queue_event_message();
	}
}



void SndMsgCmd::okCallback ( Widget, XtPointer clientData, XtPointer )
{
	SndMsgCmd * obj = ( SndMsgCmd * )clientData;
	obj->Cmd::execute();
}


void SndMsgCmd::cancelCallback ( Widget, XtPointer clientData, XtPointer )
{  
	SndMsgCmd * obj = ( SndMsgCmd * )clientData;
	obj->cancel();
}

void SndMsgCmd::cancel()
{
	XtUnmanageChild( _dialog );
}


void traversal( Widget w, XtPointer, XtPointer )
{
	XmProcessTraversal( w, XmTRAVERSE_NEXT_TAB_GROUP);
}


void checkNode( Widget w, XtPointer, XtPointer callData )
{
	char c;

	XmTextVerifyCallbackStruct *cbs = 
		( XmTextVerifyCallbackStruct * )callData;

	if (cbs->reason == XmCR_ACTIVATE) 
	{
		XmProcessTraversal(w, XmTRAVERSE_NEXT_TAB_GROUP);
		return;
	}

	if (cbs->startPos < cbs->currInsert) /* backspace */
		return;

	if (! cbs->text->ptr || (!isdigit(c = cbs->text->ptr[0])
			&& c != '.' && c != ' ' && c != ',') )
		cbs->doit = False;
}
